from django.db.models import Manager
from django.contrib.contenttypes.models import ContentType


class FieldHistoryManager(Manager):

    def get_for_model(self, object):
        content_type = ContentType.objects.get_for_model(object)
        return self.filter(object_id=object.pk,
                           content_type=content_type)

    def get_for_model_and_field(self, object, field):
        return self.get_for_model(object).filter(field_name=field)



import inspect

from django.contrib.contenttypes.models import ContentType
from django.apps import apps
from django.core import serializers
from django.core.management import BaseCommand

from field_history.models import FieldHistory
from field_history.tracker import FieldHistoryTracker, get_serializer_name


class Command(BaseCommand):

    help = "Adds initial FieldHistory objects"

    def handle(self, *args, **options):
        models = []
        for model in apps.get_models():
            for member in inspect.getmembers(model):
                if isinstance(member[1], FieldHistoryTracker):
                    models.append((model, member[1].fields))
                    break

        if models:
            self.stdout.write('Creating initial field history for {} models\n'.format(len(models)))

            for model_fields in models:
                model = model_fields[0]
                fields = model_fields[1]

                for obj in model._default_manager.all():
                    for field in list(fields):
                        content_type = ContentType.objects.get_for_model(obj)
                        if not FieldHistory.objects.filter(
                                object_id=obj.pk,
                                content_type=content_type,
                                field_name=field).exists():
                            data = serializers.serialize(get_serializer_name(),
                                                         [obj],
                                                         fields=[field])
                            FieldHistory.objects.create(
                                object=obj,
                                field_name=field,
                                serialized_data=data,
                            )
        else:
            self.stdout.write('There are no models to create field history for.')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('field_history', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='fieldhistory',
            options={'get_latest_by': 'date_created'},
        ),
        migrations.AlterField(
            model_name='fieldhistory',
            name='field_name',
            field=models.CharField(max_length=500, db_index=True),
            preserve_default=True,
        ),
    ]



from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.management import BaseCommand, CommandError

from field_history.models import FieldHistory


class Command(BaseCommand):

    help = """Updates FieldHistory field_names for a given model. Run this after renaming a model field via migrations.

Example:

    python manage.py renamefieldhistory --model=myapp.User --from_field=username to_field=handle
"""

    def add_arguments(self, parser):
        parser.add_argument(
            '--model',
            type=str,
            help='The model class to update in '
                 'app_label.model_name format (e.g. auth.User)')

        parser.add_argument(
            '--from_field',
            type=str,
            help='The old model field name')

        parser.add_argument(
            '--to_field',
            type=str,
            help='The new model field name')

    def handle(self, *args, **options):
        model_name = options.get('model')
        from_field = options.get('from_field')
        to_field = options.get('to_field')

        if not model_name:
            raise CommandError('--model_name is a required argument')
        if not from_field:
            raise CommandError('--from_field is a required argument')
        if not to_field:
            raise CommandError('--to_field is a required argument')

        model = apps.get_model(model_name)
        content_type = ContentType.objects.get_for_model(model)
        field_histories = FieldHistory.objects.filter(content_type=content_type, field_name=from_field)

        self.stdout.write('Updating {} FieldHistory object(s)\n'.format(field_histories.count()))

        field_histories.update(field_name=to_field)



"""
Serialize data to/from JSON
Only one change comparing to Django build-in json serializer: obj.local_fields -> obj.fields for supporting
nested models, parent fields are included to output
"""

# Avoid shadowing the standard library json module
from __future__ import absolute_import
from __future__ import unicode_literals

import warnings

from django.core.serializers.json import Serializer as JsonSerializer
from django.utils import six

try:
    from django.utils.deprecation import RemovedInDjango19Warning
except ImportError:
    RemovedInDjango19Warning = None


class Serializer(JsonSerializer):
    """
    Convert a queryset to JSON.
    """
    def serialize(self, queryset, **options):
        """
        Serialize a queryset.
        """
        self.options = options

        self.stream = options.pop("stream", six.StringIO())
        self.selected_fields = options.pop("fields", None)
        self.use_natural_keys = options.pop("use_natural_keys", False)
        if self.use_natural_keys and RemovedInDjango19Warning is not None:
            warnings.warn("``use_natural_keys`` is deprecated; use ``use_natural_foreign_keys`` instead.",
                          RemovedInDjango19Warning)
        self.use_natural_foreign_keys = options.pop('use_natural_foreign_keys', False) or self.use_natural_keys
        self.use_natural_primary_keys = options.pop('use_natural_primary_keys', False)

        self.start_serialization()
        self.first = True
        for obj in queryset:
            self.start_object(obj)
            # Use the concrete parent class' _meta instead of the object's _meta
            # This is to avoid local_fields problems for proxy models. Refs #17717.
            concrete_model = obj._meta.concrete_model
            # only one change local_fields -> fields for supporting nested models
            for field in concrete_model._meta.fields:
                if field.serialize:
                    if field.remote_field is None:
                        if self.selected_fields is None or field.attname in self.selected_fields:
                            self.handle_field(obj, field)
                    else:
                        if self.selected_fields is None or field.attname[:-3] in self.selected_fields:
                            self.handle_fk_field(obj, field)
            for field in concrete_model._meta.many_to_many:
                if field.serialize:
                    if self.selected_fields is None or field.attname in self.selected_fields:
                        self.handle_m2m_field(obj, field)
            self.end_object(obj)
            if self.first:
                self.first = False
        self.end_serialization()
        return self.getvalue()



from .tracker import FieldHistoryTracker
from django import VERSION


if VERSION >= (1, 10):
    from django.utils.deprecation import MiddlewareMixin

    class FieldHistoryMiddleware(MiddlewareMixin):

        def process_request(self, request):
            FieldHistoryTracker.thread.request = request
else:
    class FieldHistoryMiddleware(object):

        def process_request(self, request):
            FieldHistoryTracker.thread.request = request



# -*- coding: utf-8 -*-
from django.conf import settings
try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:  # Django < 1.9 pragma: no cover
    from django.contrib.contenttypes.generic import GenericForeignKey
from django.core import serializers
from django.db import models
from django.utils.encoding import python_2_unicode_compatible

from .managers import FieldHistoryManager

OBJECT_ID_TYPE_SETTING = 'FIELD_HISTORY_OBJECT_ID_TYPE'


def instantiate_object_id_field(object_id_class_or_tuple=models.TextField):
    """
    Instantiates and returns a model field for FieldHistory.object_id.

    object_id_class_or_tuple may be either a Django model field class or a
    tuple of (model_field, kwargs), where kwargs is a dict passed to
    model_field's constructor.
    """
    if isinstance(object_id_class_or_tuple, (list, tuple)):
        object_id_class, object_id_kwargs = object_id_class_or_tuple
    else:
        object_id_class = object_id_class_or_tuple
        object_id_kwargs = {}

    if not issubclass(object_id_class, models.fields.Field):
        raise TypeError('settings.%s must be a Django model field or (field, kwargs) tuple' % OBJECT_ID_TYPE_SETTING)
    if not isinstance(object_id_kwargs, dict):
        raise TypeError('settings.%s kwargs must be a dict' % OBJECT_ID_TYPE_SETTING)

    return object_id_class(db_index=True, **object_id_kwargs)


@python_2_unicode_compatible
class FieldHistory(models.Model):
    object_id = instantiate_object_id_field(getattr(settings, OBJECT_ID_TYPE_SETTING, models.TextField))
    content_type = models.ForeignKey('contenttypes.ContentType', db_index=True, on_delete=models.CASCADE)
    object = GenericForeignKey()
    field_name = models.CharField(max_length=500, db_index=True)
    serialized_data = models.TextField()
    date_created = models.DateTimeField(auto_now_add=True, db_index=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.CASCADE)

    objects = FieldHistoryManager()

    class Meta:
        app_label = 'field_history'
        get_latest_by = 'date_created'

    def __str__(self):
        return u'{} field history for {}'.format(self.field_name, self.object)

    @property
    def field_value(self):
        instances = serializers.deserialize('json', self.serialized_data)
        instance = list(instances)[0].object
        return getattr(instance, self.field_name)
